from btypes import *

cstring_sjis = CString('shift-jis')

#------------------------------------------------------------------------------

class Header(Struct):
    magic = ByteString(4)
    file_type = ByteString(4)
    file_size = uint32
    section_count = uint32
    unknown0 = ByteString(4)
    __padding__ = Padding(12)

#------------------------------------------------------------------------------

class StringTable:

    class Header(Struct):
        entry_count = uint16
        __padding__ = Padding(2)

    class Entry(Struct):
        string_hash = uint16
        offset = uint16

    @staticmethod
    def calculate_string_hash(string):
        h = 0
        for c in string:
            h = (h*3 + c) & 0xFFFF
        return h

    @staticmethod
    def pack(stream,strings):
        strings = [string.encode('shift-jis') for string in strings]

        header = StringTable.Header()
        header.entry_count = len(strings)
        StringTable.Header.pack(stream,header)

        offset = StringTable.Header.sizeof() + len(strings)*StringTable.Entry.sizeof()

        for string in strings:
            entry = StringTable.Entry()
            entry.string_hash = StringTable.calculate_string_hash(string)
            entry.offset = offset
            StringTable.Entry.pack(stream,entry)
            offset += len(string) + 1

        for string in strings:
            stream.write(string)
            stream.write(b'\x00')

    @staticmethod
    def unpack(stream):
        header = StringTable.Header.unpack(stream)
        stream.read(StringTable.Entry.sizeof()*header.entry_count)
        return [cstring_sjis.unpack(stream) for _ in range(header.entry_count)]

#______________________________________________________________________________
