from btypes import *

#------------------------------------------------------------------------------

class Header(Struct):
    magic = ByteString(4)
    section_size = uint32
    entry_count = uint16
    __padding__ = Padding(2)
    is_weighted_offset = uint32
    index_offset = uint32

    def __init__(self):
        self.magic = b'DRW1'


class MatrixDescriptor:

    def __init__(self,index,is_weighted):
        self.index = index
        self.is_weighted = is_weighted

#------------------------------------------------------------------------------

def pack(stream,matrix_descriptors):
    base = stream.tell()
    header = Header()
    header.entry_count = len(matrix_descriptors)
    stream.write(b'\x00'*Header.sizeof())

    header.is_weighted_offset = stream.tell() - base
    for matrix_descriptor in matrix_descriptors:
        bool8.pack(stream,matrix_descriptor.is_weighted)

    align(stream,2)
    header.index_offset = stream.tell() - base
    for matrix_descriptor in matrix_descriptors:
        uint16.pack(stream,matrix_descriptor.index)

    align(stream,0x20)
    header.section_size = stream.tell() - base
    stream.seek(base)
    Header.pack(stream,header)
    stream.seek(base + header.section_size)


def unpack(stream):
    base = stream.tell()
    header = Header.unpack(stream)
    if header.magic != b'DRW1':
        raise FormatError('invalid magic')

    matrix_descriptors = [MatrixDescriptor(None,None) for _ in range(header.entry_count)]

    stream.seek(base + header.is_weighted_offset)
    for matrix_descriptor in matrix_descriptors:
        matrix_descriptor.is_weighted = bool8.unpack(stream)

    stream.seek(base + header.index_offset)
    for matrix_descriptor in matrix_descriptors:
        matrix_descriptor.index = uint16.unpack(stream)

    stream.seek(base + header.section_size)
    return matrix_descriptors

#______________________________________________________________________________
