from btypes import *
import j3d

class FormatError(Exception): pass

#------------------------------------------------------------------------------

class Header(Struct):
    magic = ByteString(4)
    section_size = uint32
    entry_count = uint16
    __padding__ = Padding(2)
    joint_offset = uint32
    index_offset = uint32
    name_offset = uint32

    def __init__(self):
        self.magic = b'JNT1'


class Joint(Struct):
    unknown0 = uint16
    unknown1 = uint8
    __padding__ = Padding(1)
    scale_x = float32
    scale_y = float32
    scale_z = float32
    rotate_x = FixedPointConverter(sint16,180/32768)
    rotate_y = FixedPointConverter(sint16,180/32768)
    rotate_z = FixedPointConverter(sint16,180/32768)
    __padding__ = Padding(2)
    translate_x = float32
    translate_y = float32
    translate_z = float32
    unknown2 = float32
    min_x = float32
    min_y = float32
    min_z = float32
    max_x = float32
    max_y = float32
    max_z = float32

    def __init__(self,name):
        self.name = name
        self.unknown0 = 0
        self.unknown1 = 0
        self.scale_x = 1
        self.scale_y = 1
        self.scale_z = 1
        self.rotate_x = 0
        self.rotate_y = 0
        self.rotate_z = 0
        self.translate_x = 0
        self.translate_y = 0
        self.translate_z = 0

#------------------------------------------------------------------------------

def pack(stream,joints):
    base = stream.tell()
    header = Header()
    header.entry_count = len(joints)
    stream.write(b'\x00'*Header.sizeof())

    header.joint_offset = stream.tell() - base
    for joint in joints:
        Joint.pack(stream,joint)

    header.index_offset = stream.tell() - base
    for index in range(len(joints)):
        uint16.pack(stream,index)

    align(stream,4)
    header.name_offset = stream.tell() - base
    j3d.StringTable.pack(stream,(joint.name for joint in joints))

    align(stream,0x20)
    header.section_size = stream.tell() - base
    stream.seek(base)
    Header.pack(stream,header)
    stream.seek(base + header.section_size)


def unpack(stream):
    base = stream.tell()
    header = Header.unpack(stream)
    if header.magic != b'JNT1':
        raise FormatError('invalid magic')

    stream.seek(base + header.joint_offset)
    joints = [Joint.unpack(stream) for _ in range(header.entry_count)]

    stream.seek(base + header.index_offset)
    for index in range(header.entry_count):
        if index != uint16.unpack(stream):
            raise FormatError('invalid index')
    
    stream.seek(base + header.name_offset)
    names = j3d.StringTable.unpack(stream)
    for joint,name in zip(joints,names):
        joint.name = name

    stream.seek(base + header.section_size)
    return joints

#______________________________________________________________________________
