from btypes import *
import gx
import j3d

#------------------------------------------------------------------------------

class Header(Struct):
    magic = ByteString(4)
    section_size = uint32
    material_count = uint16
    __padding__ = Padding(2)
    entry_offset = uint32
    entry_index_offset = uint32
    name_offset = uint32
    indirect_entry_offset = uint32
    cullmode_offset = uint32
    matcolor_offset = uint32
    numchans_offset = uint32
    chanctrl_offset = uint32
    ambcolor_offset = uint32
    unknown0_offset = uint32
    numtexgens_offset = uint32
    texcoordgen_offset = uint32
    unknown1_offset = uint32
    texmatrix_offset = uint32
    unknown2_offset = uint32
    texture_offset = uint32
    tevorder_offset = uint32
    tevcolor_offset = uint32
    kcolor_offset = uint32
    numtevstages_offset = uint32
    tevcombiner_offset = uint32
    swapmode_offset = uint32
    swaptable_offset = uint32
    fog_offset = uint32
    alphacompare_offset = uint32
    blendmode_offset = uint32
    zmode_offset = uint32
    zcomploc_offset = uint32
    dither_offset = uint32
    unknown3_offset = uint32

    def __init__(self):
        self.magic = b'MAT3'


class Color(Struct):
    """Unsigned RGBA8 color."""
    r = uint8
    g = uint8
    b = uint8
    a = uint8

    def __init__(self,r=0x00,g=0x00,b=0x00,a=0xFF):
        self.r = r
        self.g = g
        self.b = b
        self.a = a


class ColorS16(Struct):
    """Signed RGBA16 color."""
    r = sint16
    g = sint16
    b = sint16
    a = sint16

    def __init__(self,r=0x00,g=0x00,b=0x00,a=0xFF):
        self.r = r
        self.g = g
        self.b = b
        self.a = a


class ZMode(Struct):
    """Arguments to GXSetZMode."""
    enable = bool8
    function = uint8
    update_enable = bool8
    __padding__ = Padding(1)

    def __init__(self):
        self.enable = True
        self.function = gx.LEQUAL
        self.update_enable = True


class AlphaCompare(Struct):
    """Arguments to GXSetAlphaCompare."""
    compare0 = uint8
    reference0 = uint8
    operator = uint8
    compare1 = uint8
    reference1 = uint8
    __padding__ = Padding(3)

    def __init__(self):
        self.compare0 = gx.ALWAYS
        self.reference0 = 0
        self.operator = gx.AOP_AND
        self.compare1 = gx.ALWAYS
        self.reference1 = 0


class BlendMode(Struct):
    """Arguments to GXSetBlendMode."""
    function = uint8
    source_factor = uint8
    destination_factor = uint8
    logical_operation = uint8

    def __init__(self):
        self.function = gx.BM_NONE
        self.source_factor = gx.BL_SRCALPHA
        self.destination_factor = gx.BL_INVSRCALPHA
        self.logical_operation = gx.LO_CLEAR


class ChanCtrl(Struct):
    """Arguments to GXSetChanCtrl."""
    light_enable = bool8
    material_source = uint8
    light_mask = uint8
    diffuse_function = uint8
    attenuation_function = uint8
    ambient_source = uint8
    __padding__ = Padding(2)

    def __init__(self,light_enable,ambient_source,material_source,light_mask,diffuse_function,attenuation_function):
        self.light_enable = light_enable
        self.ambient_source = ambient_source
        self.material_source = material_source
        self.light_mask = light_mask
        self.diffuse_function = diffuse_function
        self.attenuation_function = attenuation_function


class TexCoordGen(Struct):
    """Arguments to GXSetTexCoordGen."""
    function = uint8
    source = uint8
    matrix = uint8
    __padding__ = Padding(1)

    def __init__(self):
        self.function = gx.TG_MTX2x4
        self.source = gx.TG_TEX0
        self.matrix = gx.IDENTITY


class TexMatrix(Struct):
    projection = uint8
    matrix_type = uint8
    __padding__ = Padding(2)
    center_s = float32
    center_t = float32
    unknown0 = float32
    scale_s = float32
    scale_t = float32
    rotate = FixedPointConverter(sint16,180/32768)
    __padding__ = Padding(2)
    translate_s = float32
    translate_t = float32
    premtx = Array(Array(float32,4),4)

    def __init__(self):
        self.matrix_type = 0
        self.projection = gx.TG_MTX2x4
        self.center_s = 0.5
        self.center_t = 0.5
        self.unknown0 = 0.5
        self.rotate = 0
        self.scale_s = 1
        self.scale_t = 1
        self.translate_s = 0
        self.translate_t = 0
        self.premtx = [[1,0,0,0],[0,1,0,0],[0,0,1,0],[0,0,0,1]]


class TevColorIn(Struct):
    """Arguments to GXSetTevColorIn."""
    a = uint8
    b = uint8
    c = uint8
    d = uint8

    def __init__(self,a=gx.CC_ZERO,b=gx.CC_ZERO,c=gx.CC_ZERO,d=gx.CC_TEXC):
        self.a = a
        self.b = b
        self.c = c
        self.d = d


class TevAlphaIn(Struct):
    """Arguments to GXSetTevAlphaIn."""
    a = uint8
    b = uint8
    c = uint8
    d = uint8

    def __init__(self,a=gx.CA_ZERO,b=gx.CA_ZERO,c=gx.CA_ZERO,d=gx.CA_TEXA):
        self.a = a
        self.b = b
        self.c = c
        self.d = d


class TevColorOp(Struct):
    """Arguments to GXSetTevColorOp."""
    operator = uint8
    bias = uint8
    scale = uint8
    clamp = uint8
    output = uint8

    def __init__(self):
        self.operator = gx.TEV_ADD
        self.bias = gx.TB_ZERO
        self.scale = gx.CS_SCALE_1
        self.clamp = True
        self.output = gx.TEVPREV


class TevAlphaOp(Struct):
    """Arguments to GXSetTevAlphaOp."""
    operator = uint8
    bias = uint8
    scale = uint8
    clamp = uint8
    output = uint8

    def __init__(self):
        self.operator = gx.TEV_ADD
        self.bias = gx.TB_ZERO
        self.scale = gx.CS_SCALE_1
        self.clamp = True
        self.output = gx.TEVPREV


class TevCombiner(Struct):
    unknown0 = uint8
    colorin = TevColorIn
    colorop = TevColorOp
    alphain = TevAlphaIn
    alphaop = TevAlphaOp
    __padding__ = Padding(1)

    def __init__(self):
        self.unknown0 = 0xFF


class TevOrder(Struct):
    """Arguments to GXSetTevOrder."""
    texcoord = uint8
    texmap = uint8
    color = uint8
    __padding__ = Padding(1)

    def __init__(self,texcoord=gx.TEXCOORD_NULL,texmap=gx.TEXMAP_NULL,color=gx.COLOR_NULL):
        self.texcoord = texcoord
        self.texmap = texmap
        self.color = color


class TevSwapMode(Struct):
    """Arguments to GXSetTevSwapMode."""
    color_table = uint8
    texture_table = uint8
    __padding__ = Padding(2)

    def __init__(self):
        self.color_table = gx.TEV_SWAP0
        self.texture_table = gx.TEV_SWAP0


class TevSwapModeTable(Struct):
    """Arguments to GXSetTevSwapModeTable."""
    r = uint8
    g = uint8
    b = uint8
    a = uint8

    def __init__(self):
        self.r = gx.CH_RED
        self.g = gx.CH_GREEN
        self.b = gx.CH_BLUE
        self.a = gx.CH_ALPHA


class TevIndirect(Struct):
    """Arguments to GXSetTevIndirect."""
    stage = uint8
    format = uint8
    bias_selection = uint8
    matrix = uint8
    wrap_s = uint8
    wrap_t = uint8
    add_previous = bool8
    use_original_lod = bool8
    bump_alpha_selection = uint8
    __padding__ = Padding(3)

    def __init__(self):
        self.stage = gx.INDTEXSTAGE0
        self.format = gx.ITF_8
        self.bias_selection = gx.ITB_NONE
        self.matrix = gx.ITM_OFF
        self.wrap_s = gx.ITW_OFF
        self.wrap_t = gx.ITW_OFF
        self.add_previous = False
        self.use_original_lod = False
        self.bump_alpha_selection = gx.ITBA_OFF


class IndTexOrder(Struct):
    """Arguments to GXSetIndTexOrder."""
    texcoord = uint8
    texmap = uint8
    __padding__ = Padding(2)

    def __init__(self):
        self.texcoord = gx.TEXCOORD_NULL
        self.texmap = gx.TEXMAP_NULL


class IndTexCoordScale(Struct):
    """Arguments to GXSetIndTexCoordScale."""
    scale_s = uint8
    scale_t = uint8
    __padding__ = Padding(2)

    def __init__(self):
        self.scale_s = gx.ITS_1
        self.scale_t = gx.ITS_1


class IndTexMatrix(Struct):
    """Arguments to GXSetIndTexMatrix."""
    offset_matrix = Array(Array(float32,3),2)
    scale_exponent = sint8
    __padding__ = Padding(3)

    def __init__(self):
        self.offset_matrix = [[0.5,0,0],[0,0.5,0]]
        self.scale_exponent = 1


class Fog(Struct):
    """Arguments to GXSetFog and GXSetFogRangeAdj."""
    function = uint8
    enable = bool8
    center = uint16
    start_z = float32
    end_z = float32
    near_z = float32
    far_z = float32
    color = Color
    table = Array(uint16,10)

    def __init__(self):
        self.function = gx.FOG_NONE
        self.start_z = 0
        self.end_z = 0
        self.near_z = 0
        self.far_z = 0
        self.color = Color(0xFF,0xFF,0xFF)

        self.enable = False
        self.center = 0
        self.table = [0]*10


class Entry(Struct):
    unknown0 = uint8 # related to transparency sorting
    cullmode_index = uint8
    numchans_index = uint8
    numtexgens_index = uint8
    numtevstages_index = uint8
    zcomploc_index = uint8
    zmode_index = uint8
    dither_index = uint8
    matcolor_indices = Array(uint16,2)
    chanctrl_indices = Array(uint16,4) # GX_COLOR0,GX_ALPHA0,GX_COLOR1,GX_ALPHA1
    ambcolor_indices = Array(uint16,2)
    unknown1 = Array(uint16,8)
    texcoordgen_indices = Array(uint16,8)
    unknown2 = Array(uint16,8)
    texmatrix_indices = Array(uint16,10)
    unknown3 = Array(uint16,20)
    texture_indices = Array(uint16,8)
    kcolor_indices = Array(uint16,4)
    kcolorsels = Array(uint8,16)
    kalphasels = Array(uint8,16)
    tevorder_indices = Array(uint16,16)
    tevcolor_indices = Array(uint16,4) # GX_TEVREG0,GX_TEVREG1,GX_TEVREG2,GX_TEVPREV
    tevcombiner_indices = Array(uint16,16)
    swapmode_indices = Array(uint16,16)
    swaptable_indices = Array(uint16,4)
    unknown4 = Array(uint16,12)
    fog_index = uint16
    alphacompare_index = uint16
    blendmode_index = uint16
    unknown5 = uint16

    def __init__(self):
        self.unknown0 = 1
        self.cullmode_index = 0xFF
        self.numchans_index = 0xFF
        self.numtexgens_index = 0xFF
        self.numtevstages_index = 0xFF
        self.zcomploc_index = 0xFF
        self.zmode_index = 0xFF
        self.dither_index = 0xFF
        self.matcolor_indices = [0xFFFF]*2
        self.chanctrl_indices = [0xFFFF]*4
        self.ambcolor_indices = [0xFFFF]*2
        self.unknown1 = [0xFFFF]*8
        self.texcoordgen_indices = [0xFFFF]*8
        self.unknown2 = [0xFFFF]*8
        self.texmatrix_indices = [0xFFFF]*10
        self.unknown3 = [0xFFFF]*20
        self.texture_indices = [0xFFFF]*8
        self.kcolor_indices = [0xFFFF]*4
        self.kcolorsels = [gx.TEV_KCSEL_K0]*16
        self.kalphasels = [gx.TEV_KASEL_K0_A]*16
        self.tevorder_indices = [0xFFFF]*16
        self.tevcolor_indices = [0xFFFF]*4
        self.tevcombiner_indices = [0xFFFF]*16
        self.swapmode_indices = [0xFFFF]*16
        self.swaptable_indices = [0xFFFF]*4
        self.unknown4 = [0xFFFF]*12
        self.fog_index = 0xFFFF
        self.alphacompare_index = 0xFFFF
        self.blendmode_index = 0xFFFF
        self.unknown5 = 0


class IndirectEntry(Struct):
    unknown0 = uint8 # enable or numindstages?
    unknown1 = uint8 # enable or numindstages?
    __padding__ = Padding(2)
    indtexorders = Array(IndTexOrder,4)
    indtexmatrices = Array(IndTexMatrix,3)
    indtexcoordscales = Array(IndTexCoordScale,4)
    tevindirects = Array(TevIndirect,16)

    def __init__(self):
        self.unknown0 = 0
        self.unknown1 = 0
        self.indtexorders = [IndTexOrder()]*4
        self.indtexcoordscales = [IndTexCoordScale()]*4
        self.indtexmatrices = [IndTexMatrix()]*3
        self.tevindirects = [TevIndirect()]*16

#------------------------------------------------------------------------------

def pack(stream,meshes):
    base = stream.tell()
    header = Header()
    header.material_count = len(meshes)
    header.unknown0_offset = 0
    header.unknown1_offset = 0
    header.unknown2_offset = 0
    stream.write(b'\x00'*Header.sizeof())

    header.entry_offset = stream.tell() - base

    matcolors = [Color(0xFF,0xFF,0xFF)]
    textures = []

    for mesh in meshes:
        entry = Entry()

        entry.cullmode_index = 0
        entry.numchans_index = 0
        entry.numtevstages_index = 0
        entry.zcomploc_index = 1
        entry.zmode_index = 0
        entry.alphacompare_index = 0
        entry.blendmode_index = 0
        entry.dither_index = 0
        entry.fog_index = 0
        entry.ambcolor_indices = [0,0]
        entry.swapmode_indices[0] = 0
        entry.tevcolor_indices = [0,0,0,0]
        entry.kcolor_indices = [0,0,0,0]
        entry.swaptable_indices = [0,0,0,0]

        entry.matcolor_indices = [len(matcolors),0]
        matcolors.append(mesh.material_color)

        if mesh.has_colors:
            entry.chanctrl_indices = [1,0,0,0]
        else:
            entry.chanctrl_indices = [0,0,0,0]

        if mesh.texture_index is None:
            entry.numtexgens_index = 0
            entry.tevorder_indices[0] = 0
            entry.tevcombiner_indices[0] = 0
        else:
            entry.numtexgens_index = 1
            entry.texcoordgen_indices[0] = 0
            entry.tevorder_indices[0] = 1
            entry.tevcombiner_indices[0] = 1

            entry.texture_indices[0] = len(textures)
            textures.append(mesh.texture_index)

        Entry.pack(stream,entry)

    header.entry_index_offset = stream.tell() - base
    for index in range(len(meshes)):
        uint16.pack(stream,index)

    align(stream,4)
    header.name_offset = stream.tell() - base
    j3d.StringTable.pack(stream,(mesh.material_name for mesh in meshes))

    align(stream,4)
    header.indirect_entry_offset = stream.tell() - base
    for _ in range(len(meshes)):
        IndirectEntry.pack(stream,IndirectEntry())

    header.cullmode_offset = stream.tell() - base
    uint32.pack(stream,gx.CULL_BACK)

    header.matcolor_offset = stream.tell() - base
    for color in matcolors:
        Color.pack(stream,Color(color.r,color.g,color.b,0xFF))

    header.numchans_offset = stream.tell() - base
    uint8.pack(stream,1)

    align(stream,4)
    header.chanctrl_offset = stream.tell() - base
    ChanCtrl.pack(stream,ChanCtrl(False,gx.SRC_REG,gx.SRC_REG,gx.LIGHT_NULL,gx.DF_NONE,gx.AF_NONE))
    ChanCtrl.pack(stream,ChanCtrl(False,gx.SRC_REG,gx.SRC_VTX,gx.LIGHT_NULL,gx.DF_NONE,gx.AF_NONE))

    header.ambcolor_offset = stream.tell() - base
    Color.pack(stream,Color(0,0,0,0))

    header.numtexgens_offset = stream.tell() - base
    uint8.pack(stream,0)
    uint8.pack(stream,1)

    header.unknown0_offset = header.numtexgens_offset

    align(stream,4)
    header.texcoordgen_offset = stream.tell() - base
    TexCoordGen.pack(stream,TexCoordGen())

    header.texmatrix_offset = 0

    header.texture_offset = stream.tell() - base
    for texture in textures:
        uint16.pack(stream,texture)

    align(stream,4)
    header.tevorder_offset = stream.tell() - base
    TevOrder.pack(stream,TevOrder(gx.TEXCOORD_NULL,gx.TEXMAP_NULL,gx.COLOR0A0))
    TevOrder.pack(stream,TevOrder(gx.TEXCOORD0,gx.TEXMAP0,gx.COLOR0A0))

    header.tevcolor_offset = stream.tell() - base
    ColorS16.pack(stream,ColorS16(0xFF,0xFF,0xFF,0xFF))

    header.kcolor_offset = stream.tell() - base
    Color.pack(stream,Color(0xFF,0xFF,0xFF,0xFF))

    header.numtevstages_offset = stream.tell() - base
    uint8.pack(stream,1)

    align(stream,4)
    header.tevcombiner_offset = stream.tell() - base

    tevcombiner = TevCombiner()
    tevcombiner.colorin = TevColorIn(gx.CC_ZERO,gx.CC_ZERO,gx.CC_ZERO,gx.CC_RASC)
    tevcombiner.colorop = TevColorOp()
    tevcombiner.alphain = TevAlphaIn(gx.CA_ZERO,gx.CA_ZERO,gx.CA_ZERO,gx.CA_RASA)
    tevcombiner.alphaop = TevAlphaOp()
    TevCombiner.pack(stream,tevcombiner)

    tevcombiner = TevCombiner()
    tevcombiner.colorin = TevColorIn(gx.CC_ZERO,gx.CC_TEXC,gx.CC_RASC,gx.CC_ZERO)
    tevcombiner.colorop = TevColorOp()
    tevcombiner.alphain = TevAlphaIn(gx.CA_ZERO,gx.CA_TEXA,gx.CA_RASA,gx.CA_ZERO)
    tevcombiner.alphaop = TevAlphaOp()
    TevCombiner.pack(stream,tevcombiner)

    header.swapmode_offset = stream.tell() - base
    TevSwapMode.pack(stream,TevSwapMode())

    header.swaptable_offset = stream.tell() - base
    TevSwapModeTable.pack(stream,TevSwapModeTable())

    header.fog_offset = stream.tell() - base
    Fog.pack(stream,Fog())

    header.alphacompare_offset = stream.tell() - base
    AlphaCompare.pack(stream,AlphaCompare())

    header.blendmode_offset = stream.tell() - base
    BlendMode.pack(stream,BlendMode())

    header.zmode_offset = stream.tell() - base
    ZMode.pack(stream,ZMode())

    header.zcomploc_offset = stream.tell() - base
    bool8.pack(stream,False)
    bool8.pack(stream,True)

    align(stream,4)
    header.dither_offset = stream.tell() - base
    bool8.pack(stream,False)
    bool8.pack(stream,True)

    align(stream,8)
    header.unknown3_offset = stream.tell() - base
    stream.write(b'\x00\xFF\xFF\xFF\x3F\x80\x00\x00\x3F\x80\x00\x00\x3F\x80\x00\x00')

    align(stream,0x20)
    header.section_size = stream.tell() - base
    stream.seek(base)
    Header.pack(stream,header)
    stream.seek(base + header.section_size)

#______________________________________________________________________________
