from btypes import *
import gx
import j3d

#------------------------------------------------------------------------------

class Header(Struct):
    magic = ByteString(4)
    section_size = uint32
    entry_count = uint16
    __padding__ = Padding(2)
    packet_offset = uint32
    subpacket_location_offset = uint32
    mtxidx_offset = uint32
    unknown0_offset = uint32
    index_offset = uint32
    name_offset = uint32

    def __init__(self):
        self.magic = b'MDL3'


class PacketLocation(Struct):
    offset = uint32
    size = uint32

    def __init__(self,offset,size):
        self.offset = offset
        self.size = size


class SubpacketLocation(Struct):
    chancolor_offset = uint16
    chanctrl_offset = uint16
    texgen_offset = uint16
    texture_offset = uint16
    tev_offset = uint16
    pixel_offset = uint16
    __padding__ = Padding(4)

#------------------------------------------------------------------------------

def pack_packet(stream,mesh,textures):

    def pack_bp_command(register,value):
        stream.write(b'\x61')
        uint32.pack(stream,(register << 24) | value)

    def pack_xf_command(register,values):
        stream.write(b'\x10')
        uint16.pack(stream,len(values) - 1)
        uint16.pack(stream,register)
        for value in values:
            uint32.pack(stream,value)

    if mesh.texture_index is None:
        pack_bp_command(0x28,0x000007)
        pack_bp_command(0xFE,0x03FFFF)
        pack_bp_command(0x3E,0)
        pack_bp_command(0x3F,0)
        pack_bp_command(0xFE,0x03FFFF)
        pack_bp_command(0x3E,0)
        pack_bp_command(0x3F,0)
    else:
        texture = textures[mesh.texture_index]
        pack_bp_command(0x94,mesh.texture_index)
        pack_bp_command(0x88,(texture.width - 1) | ((texture.height - 1) << 10) | (texture.image_format << 20))
        pack_bp_command(0x80,0x000195)
        pack_bp_command(0x84,0x000000)
        pack_bp_command(0x28,0x000040)
        pack_bp_command(0xFE,0x03FFFF)
        pack_bp_command(0x30,texture.width - 1)
        pack_bp_command(0x31,texture.height - 1)
        pack_bp_command(0xFE,0x03FFFF)
        pack_bp_command(0x3E,0)
        pack_bp_command(0x3F,0)

    for i in (gx.TEVREG0,gx.TEVREG1,gx.TEVREG2):
        pack_bp_command(0xE0 + 2*i,0x0FF0FF)
        pack_bp_command(0xE1 + 2*i,0x0FF0FF)
        pack_bp_command(0xE1 + 2*i,0x0FF0FF)
        pack_bp_command(0xE1 + 2*i,0x0FF0FF)

    for i in range(4):
        pack_bp_command(0xE0 + 2*i,0x8FF0FF)
        pack_bp_command(0xE1 + 2*i,0x8FF0FF)

    if mesh.texture_index is None:
        pack_bp_command(0xC0,0x08FFFA)
        pack_bp_command(0xC1,0x08FFD0)
        pack_bp_command(0x10,0x000000)
    else:
        pack_bp_command(0xC0,0x08F8AF)
        pack_bp_command(0xC1,0x08F2F0)
        pack_bp_command(0x10,0x000000)

    pack_bp_command(0xF6,0xE338C4)
    pack_bp_command(0xF7,0xE338CE)
    pack_bp_command(0xF8,0xE338C4)
    pack_bp_command(0xF9,0xE338CE)
    pack_bp_command(0xFA,0xE338C4)
    pack_bp_command(0xFB,0xE338CE)
    pack_bp_command(0xFC,0xE338C4)
    pack_bp_command(0xFD,0xE338CE)

    for i in range(4):
        pack_bp_command(0xFE,0x03FFFF)
        pack_bp_command(0x2E,0x000000)
        pack_bp_command(0x2F,0x000000)

    pack_bp_command(0x27,0xFFFFFF)
    pack_bp_command(0x0F,0x000000)

    pack_bp_command(0xEE,0x000000)
    pack_bp_command(0xEF,0x40000F)
    pack_bp_command(0xF0,0x000001)
    pack_bp_command(0xF1,0x000000)
    pack_bp_command(0xF2,0xFFFFFF)
    pack_bp_command(0xE8,0x000156)

    pack_bp_command(0xF3,0x3F0000)
    pack_bp_command(0xFE,0x00FFE7)
    pack_bp_command(0x41,0x0004A0)
    pack_bp_command(0x40,0x000017)
    pack_bp_command(0xFE,0x000040)
    pack_bp_command(0x43,0x000040)
    pack_bp_command(0xFE,0x07FC3F)

    if mesh.texture_index is None:
        pack_bp_command(0x00,0x004010)
    else:
        pack_bp_command(0x00,0x004011)

    if mesh.texture_index is not None:
        pack_xf_command(0x1040,[0x00000280])
        pack_xf_command(0x1050,[0x0000003D])

    pack_xf_command(0x100C,[0xFF | (mesh.material_color.b << 8) | (mesh.material_color.g << 16) | (mesh.material_color.r << 24),0xFFFFFFFF])
    pack_xf_command(0x100A,[0x00000000,0x00000000])

    if mesh.has_colors:
        pack_xf_command(0x100E,[0x00000401,0x00000400,0x00000400,0x00000400])
    else:
        pack_xf_command(0x100E,[0x00000400,0x00000400,0x00000400,0x00000400])

    pack_xf_command(0x1009,[1])

    if mesh.texture_index is None:
        pack_xf_command(0x103F,[0])
    else:
        pack_xf_command(0x103F,[1])

    align(stream,0x20,b'\x00')


def pack(stream,meshes,textures):
    base = stream.tell()
    header = Header()
    header.entry_count = len(meshes)
    stream.write(b'\x00'*Header.sizeof())

    align(stream,0x20)
    header.packet_offset = stream.tell() - base
    stream.write(b'\x00'*len(meshes)*PacketLocation.sizeof())
    align(stream,0x20)
    offsets = [stream.tell()]

    for mesh in meshes:
        pack_packet(stream,mesh,textures)
        offsets.append(stream.tell())

    stream.seek(base + header.packet_offset)
    for i in range(len(meshes)):
        PacketLocation.pack(stream,PacketLocation(offsets[i] - stream.tell(),offsets[i + 1] - offsets[i]))

    stream.seek(offsets[-1])

    header.subpacket_location_offset = stream.tell() - base

    for mesh in meshes:
        subpacket_location = SubpacketLocation()

        if mesh.texture_index is None:
            subpacket_location.chancolor_offset = 330
            subpacket_location.chanctrl_offset = 356
            subpacket_location.texgen_offset = 330
            subpacket_location.texture_offset = 0
            subpacket_location.tev_offset = 35
            subpacket_location.pixel_offset = 260
        else:
            subpacket_location.chancolor_offset = 368
            subpacket_location.chanctrl_offset = 394
            subpacket_location.texgen_offset = 350
            subpacket_location.texture_offset = 0
            subpacket_location.tev_offset = 55
            subpacket_location.pixel_offset = 280

        SubpacketLocation.pack(stream,subpacket_location)

    header.mtxidx_offset = stream.tell() - base
    for _ in range(len(meshes)):
        uint32.pack(stream,0x3CF3CF00)
        uint32.pack(stream,0x00F3CF3C)

    header.unknown0_offset = stream.tell() - base
    for _ in range(len(meshes)):
        uint8.pack(stream,1)

    align(stream,4)
    header.index_offset = stream.tell() - base
    for index in range(len(meshes)):
        uint16.pack(stream,index)

    align(stream,4)
    header.name_offset = stream.tell() - base
    j3d.StringTable.pack(stream,(mesh.material_name for mesh in meshes))

    align(stream,0x20)
    header.section_size = stream.tell() - base
    stream.seek(base)
    Header.pack(stream,header)
    stream.seek(base + header.section_size)

#______________________________________________________________________________
