import os.path
from btypes import *
import gx
import j3d
from PyQt4.QtGui import QImage

#------------------------------------------------------------------------------

class Header(Struct):
    magic = ByteString(4)
    section_size = uint32
    entry_count = uint16
    __padding__ = Padding(2)
    texture_offset = uint32
    name_offset = uint32

    def __init__(self):
        self.magic = b'TEX1'


class Texture(Struct):
    image_format = uint8
    unknown0 = uint8
    width = uint16
    height = uint16
    wrap_s = uint8
    wrap_t = uint8
    unknown1 = uint8
    tlut_format = uint8
    tlut_entry_count = uint16
    tlut_offset = uint32
    unknown2 = uint32
    min_filter = uint8
    mag_filter = uint8
    min_lod = FixedPointConverter(sint8,1/8)
    max_lod = FixedPointConverter(sint8,1/8)
    image_count = uint8
    unknown3 = uint8
    lod_bias = FixedPointConverter(sint16,1/100)
    image_offset = uint32

    def __init__(self,filename):
        self.image = QImage(filename)

        if self.image.isNull():
            raise Exception('could not load image {}'.format(filename))

        self.name = os.path.splitext(os.path.basename(filename))[0]
        self.unknown0 = 0
        self.unknown1 = 0
        self.unknown2 = 0
        self.unknown3 = 0
        self.image_format = gx.TF_RGBA8
        self.width = self.image.width()
        self.height = self.image.height()
        self.image_count = 1
        self.tlut_format = gx.TL_IA8
        self.tlut_entry_count = 0
        self.wrap_s = gx.REPEAT
        self.wrap_t = gx.REPEAT
        self.min_filter = gx.LINEAR
        self.mag_filter = gx.LINEAR
        self.min_lod = 0
        self.max_lod = 0
        self.lod_bias = 0

#------------------------------------------------------------------------------

def pack_image(stream,image):
    for tile_row in range(0,image.height(),4):
      for tile_col in range(0,image.width(),4):
        ar_tile = bytearray(0x20)
        gb_tile = bytearray(0x20)

        for i,row in zip(range(4),range(tile_row,image.height())):
          for j,col in zip(range(4),range(tile_col,image.width())):
            pixel = image.pixel(col,row)
            ar_tile[2*(4*i + j)] = pixel >> 24
            ar_tile[2*(4*i + j) + 1] = (pixel >> 16) & 0xFF
            gb_tile[2*(4*i + j)] = (pixel >> 8) & 0xFF
            gb_tile[2*(4*i + j) + 1] = pixel & 0xFF

        stream.write(ar_tile)
        stream.write(gb_tile)


def pack(stream,textures):
    base = stream.tell()
    header = Header()
    header.entry_count = len(textures)
    stream.write(b'\x00'*Header.sizeof())

    align(stream,0x20)
    header.texture_offset = stream.tell() - base

    offset = header.entry_count*Texture.sizeof()

    for texture in textures:
        texture.image_offset = offset
        texture.tlut_offset = 0
        Texture.pack(stream,texture)

        aligned_width = 4*((texture.width + 3)//4)
        aligned_height = 4*((texture.height + 3)//4)
        offset += 4*aligned_width*aligned_height - Texture.sizeof()

    for texture in textures:
        pack_image(stream,texture.image)

    header.name_offset = stream.tell() - base
    j3d.StringTable.pack(stream,(texture.name for texture in textures))

    align(stream,0x20)
    header.section_size = stream.tell() - base
    stream.seek(base)
    Header.pack(stream,header)
    stream.seek(base + header.section_size)

#______________________________________________________________________________
