from btypes import *
import gx

#------------------------------------------------------------------------------

class Header(Struct):
    magic = ByteString(4)
    section_size = uint32
    attribute_format_offset = uint32
    position_offset = uint32
    normal_offset = uint32
    unknown0_offset = uint32 # nbt offset?
    color_offsets = Array(uint32,2)
    texcoord_offsets = Array(uint32,8)

    def __init__(self):
        self.magic = b'VTX1'


class AttributeFormat(Struct):
    """ Arguments to GXSetVtxAttrFmt."""
    attribute = uint32
    component_count = uint32
    component_type = uint32
    scale_exponent = uint8
    __padding__ = Padding(3)

    def __init__(self,attribute,component_count,component_type,scale_exponent):
        self.attribute = attribute
        self.component_count = component_count
        self.component_type = component_type
        self.scale_exponent = scale_exponent

#------------------------------------------------------------------------------

def pack(stream,positions,colors,texcoords):
    base = stream.tell()
    header = Header()
    header.normal_offset = 0
    header.unknown0_offset = 0
    header.color_offsets = [0]*2
    header.texcoord_offsets = [0]*8
    stream.write(b'\x00'*Header.sizeof())

    header.attribute_format_offset = stream.tell() - base
    AttributeFormat.pack(stream,AttributeFormat(gx.VA_POS,gx.POS_XYZ,gx.F32,0))
    AttributeFormat.pack(stream,AttributeFormat(gx.VA_CLR0,gx.CLR_RGB,gx.RGB8,0))
    AttributeFormat.pack(stream,AttributeFormat(gx.VA_TEX0,gx.TEX_ST,gx.F32,0))
    AttributeFormat.pack(stream,AttributeFormat(gx.VA_NULL,1,0,0))

    align(stream,0x20)
    header.position_offset = stream.tell() - base
    for position in positions:
        float32.pack(stream,position.x)
        float32.pack(stream,position.y)
        float32.pack(stream,position.z)

    align(stream,0x20)
    header.color_offsets[0] = stream.tell() - base
    for color in colors:
        uint8.pack(stream,color.r)
        uint8.pack(stream,color.g)
        uint8.pack(stream,color.b)
        uint8.pack(stream,0xFF) #TODO: why is this needed?

    align(stream,0x20)
    header.texcoord_offsets[0] = stream.tell() - base
    for texcoord in texcoords:
        float32.pack(stream,texcoord.s)
        float32.pack(stream,texcoord.t)

    align(stream,0x20)
    header.section_size = stream.tell() - base
    stream.seek(base)
    Header.pack(stream,header)
    stream.seek(base + header.section_size)

#______________________________________________________________________________
